# -*- coding: utf-8 -*-

"""RPPM2 Policy Evaluation Algorithm"""
# Required imports
import argparse
import queue
import copy
import sys
from debugutils import debugprint
from PathCondition import PathCondition
from Action import Action
from Rule import Rule
from Request import Request
from NodeType import NodeType
from VisitedNode import VisitedNode
from Edge import Edge
from OrderedSet import OrderedSet

program_version = "2.0 (Copyright 2015)"

s = None
r = None
vertices = None
edges = None
u = None
v = None
pi = None


def main():
	"""RPPM2 Policy Evaluation Algorithm\nAuthor: Ishan Mehta, based on code by James Sellwood"""
	program_description = "RPPM2 Policy Evaluation Algorithm\nAuthor: Ishan Mehta, based on code by James Sellwood"
	parser = argparse.ArgumentParser(description=program_description)
	parser.add_argument('--version', action='version', version=program_version)

	test_case_no = sys.argv[1]

	# ----------------------test_case1
	if test_case_no == "1":
		req = Request("Samrat", "Scabies_Item1", "read")
		action = Action("read", "Clinician", "Item")
		setupCondition1()
		r = Rule("Clinician", action, condition)

		# This setups the system graph as shown in the diagram in the document
		setupSystemGraph1()

	# ----------------------test_case2
	elif test_case_no == "2":
		req = Request("Samrat", "Scabies_Item1", "read")
		action = Action("read", "Clinician", "Item")
		setupCondition2()
		r = Rule("Clinician", action, condition)
		setupSystemGraph2()

	# ----------------------test_case3
	elif test_case_no == "3":
		req = Request("Samrat", "Scabies_Item1", "read")
		action = Action("read", "Clinician", "Item")
		setupCondition3()
		r = Rule("Clinician", action, condition)
		setupSystemGraph3()

	# ----------------------test_case4
	elif test_case_no == "4":
		req = Request("Samrat", "Scabies_Item1", "read")
		action = Action("read", "Clinician", "Item")
		setupCondition4()
		r = Rule("Clinician", action, condition)
		setupSystemGraph4()

	# ----------------------test_case5
	elif test_case_no == "5":
		req = Request("Samrat", "Scabies_Item1", "read")
		action = Action("read", "Clinician", "Item")
		setupCondition5()
		r = Rule("Clinician", action, condition)
		setupSystemGraph5()

	# ---------------------test_case6
	elif test_case_no == "6":
		req = Request("Samrat", "Scabies_Med_Rep", "read")
		action = Action("read", "Clinician", "Item")
		setupCondition6()
		r = Rule("Clinician", action, condition)
		setupSystemGraph6()

	# ---------------------test_case7
	elif test_case_no == "7":
		req = Request("Samrat", "Scabies_Item1", "read")
		action = Action("read", "Clinician", "Item")
		setupCondition7()
		r = Rule("Clinician", action, condition)
		setupSystemGraph7()

	# ---------------------test_case8
	elif test_case_no == "8":
		req = Request("Samrat", "Scabies_Item1", "read")
		action = Action("read", "Clinician", "Item")
		setupCondition8()
		r = Rule("Clinician", action, condition)
		setupSystemGraph8()

	# ---------------------test_case9
	elif test_case_no == "9":
		req = Request("Samrat", "Scabies_Item1", "read")
		action = Action("read", "Clinician", "Item")
		setupCondition9()
		r = Rule("Clinician", action, condition)
		setupSystemGraph9()

	# ---------------------test_case10
	if test_case_no == "10":
		req = Request("a", "1,2", "read")
		action = Action("read", " ", "X,Y")
		setupCondition10()
		r = Rule("S", action, condition)

		# This setups the system graph as shown in the diagram in the document under TestCases developed by Scott
		setupSystemGraph10()

	# ---------------------test_case11
	if test_case_no == "11":
		req = Request("a", "1,2", "read")
		action = Action("read", " ", "X,Y")
		setupCondition11()
		r = Rule("S", action, condition)

		# This setups the system graph as shown in the diagram in the document under TestCases developed by Scott
		setupSystemGraph10()

	# ---------------------test_case12
	if test_case_no == "12":
		req = Request("a", "1,2", "read")
		action = Action("read", " ", "X,Y")
		setupCondition12()
		r = Rule("S", action, condition)

		# This setups the system graph as shown in the diagram in the document under TestCases developed by Scott
		setupSystemGraph10()

	# ---------------------test_case13
	if test_case_no == "13":
		req = Request("a", "1,2", "read")
		action = Action("read", " ", "X,Y")
		setupCondition13()
		r = Rule("S", action, condition)

		# This setups the system graph as shown in the diagram in the document under TestCases developed by Scott
		setupSystemGraph10()

	# ---------------------test_case14
	if test_case_no == "14":
		req = Request("a", "1,2", "read")
		action = Action("read", " ", "X,Y")
		setupCondition14()
		r = Rule("S", action, condition)

		# This setups the system graph as shown in the diagram in the document under TestCases developed by Scott
		setupSystemGraph10()

	# ---------------------test_case15
	if test_case_no == "15":
		req = Request("a", "1,2", "read")
		action = Action("read", " ", "X,Y")
		setupCondition15()
		r = Rule("S", action, condition)

		# This setups the system graph as shown in the diagram in the document under TestCases developed by Scott
		setupSystemGraph10()

	# ---------------------test_case16
	if test_case_no == "16":
		req = Request("a", "1,2", "read")
		action = Action("read", " ", "X,Y")
		setupCondition16()
		r = Rule("S", action, condition)

		# This setups the system graph as shown in the diagram in the document under TestCases developed by Scott
		setupSystemGraph10()

	# ---------------------test_case17
	if test_case_no == "17":
		req = Request("a", "1,2", "read")
		action = Action("read", " ", "X,Y")
		setupCondition17()
		r = Rule("S", action, condition)

		# This setups the system graph as shown in the diagram in the document under TestCases developed by Scott
		setupSystemGraph10()

	# ---------------------test_case18
	if test_case_no == "18":
		req = Request("a", "1,2", "read")
		action = Action("read", " ", "X,Y")
		setupCondition18()
		r = Rule("S", action, condition)

		# This setups the system graph as shown in the diagram in the document under TestCases developed by Scott
		setupSystemGraph10()

	# ---------------------test_case19
	if test_case_no == "19":
		req = Request("a", "1,2", "read")
		action = Action("read", " ", "X,Y")
		setupCondition19()
		r = Rule("S", action, condition)

		# This setups the system graph as shown in the diagram in the document under TestCases developed by Scott
		setupSystemGraph10()

	# ---------------------test_case20
	if test_case_no == "20":
		req = Request("a", "1,2", "read")
		action = Action("read", " ", "X,Y")
		setupCondition20()
		r = Rule("S", action, condition)

		# This setups the system graph as shown in the diagram in the document under TestCases developed by Scott
		setupSystemGraph10()

	# ---------------------test_case21
	if test_case_no == "21":
		req = Request("a", "1,2", "read")
		action = Action("read", " ", "X,Y")
		setupCondition21()
		r = Rule("S", action, condition)

		# This setups the system graph as shown in the diagram in the document under TestCases developed by Scott
		setupSystemGraph10()

	# Algorithm
	print("Rule Matches the request :" + str(evaluateRule(r, req)))

def setupCondition1():
	# path expresion
	pc_t = PathCondition(None, None, "treating", NodeType.relation, "Faculty")
	pc_hr = PathCondition(None, None, "healthRec", NodeType.relation, None)
	pc_c = PathCondition(None, None, "contains", NodeType.relation, None)

	pc_m = PathCondition(None, None, "member", NodeType.relation, None)
	pc_tall = PathCondition(None, None, "treating", NodeType.relation, "Faculty")

	pc_cP = PathCondition(pc_c.copy(), None, None, NodeType.plus, None)
	pc_tpc_hr = PathCondition(pc_t.copy(), pc_hr.copy(), None, NodeType.concatenate, None)

	global pc_tpc_hrpc_cP
	pc_tpc_hrpc_cP = PathCondition(pc_tpc_hr.copy(), pc_cP.copy(), None, NodeType.concatenate, None,  "Clinician",
								   None, "Item", None)

	pc_mpc_tall = PathCondition(pc_m.copy(), pc_tall.copy(), None, NodeType.concatenate, None)
	pc_hrpc_cP = PathCondition(pc_hr.copy(), pc_cP.copy(), None, NodeType.concatenate, None)

	global pc_mpc_tallpc_hrpc_cP
	pc_mpc_tallpc_hrpc_cP = PathCondition(pc_mpc_tall.copy(), pc_hrpc_cP.copy(), None, NodeType.concatenate,  None,
										  "Clinician", None, "Item", None)

	# This is the Condition of the rule containing 2 path conditions
	# 1: Clinician . treating(Faculty);healthRec;contains+ . Item
	# 2: Clinician . member;treating(Faculty);healthRec;contains+ . Item



	global condition
	condition = []
	condition.append(pc_tpc_hrpc_cP)
	condition.append(pc_mpc_tallpc_hrpc_cP)


def setupCondition2():
	# path expresion
	pc_t = PathCondition(None, None, "treating", NodeType.relation, "Faculty")
	pc_hr = PathCondition(None, None, "healthRec", NodeType.relation, None)
	pc_c = PathCondition(None, None, "contains", NodeType.relation, None)

	pc_m = PathCondition(None, None, "member", NodeType.relation, None)
	pc_tall = PathCondition(None, None, "treating", NodeType.relation, "Faculty")

	pc_cP = PathCondition(pc_c.copy(), None, None, NodeType.plus, None)
	pc_tpc_hr = PathCondition(pc_t.copy(), pc_hr.copy(), None, NodeType.concatenate, None)

	global pc_tpc_hrpc_cP
	pc_tpc_hrpc_cP = PathCondition(pc_tpc_hr.copy(), pc_cP.copy(), None, NodeType.concatenate, None, "Clinician",
								   None, "Item", None)

	pc_mpc_tall = PathCondition(pc_m.copy(), pc_tall.copy(), None, NodeType.concatenate, None)
	pc_hrpc_cP = PathCondition(pc_hr.copy(), pc_cP.copy(), None, NodeType.concatenate, None)

	global pc_mpc_tallpc_hrpc_cP
	pc_mpc_tallpc_hrpc_cP = PathCondition(pc_mpc_tall.copy(), pc_hrpc_cP.copy(), None, NodeType.concatenate, None,
										  "Clinician", None, "Item", None)

	# This is the Condition of the rule containing 2 path conditions
	# 1: Clinician . treating(Faculty);healthRec;contains+ . Item
	# 2: Clinician . member;treating(Faculty);healthRec;contains+ . Item

	global condition
	condition = []
	condition.append(pc_tpc_hrpc_cP)
	condition.append(pc_mpc_tallpc_hrpc_cP)


def setupCondition3():
	# path expresion
	pc_t = PathCondition(None, None, "treating", NodeType.relation, "Faculty")
	pc_hr = PathCondition(None, None, "healthRec", NodeType.relation, None)
	pc_c = PathCondition(None, None, "contains", NodeType.relation, None)

	pc_m = PathCondition(None, None, "member", NodeType.relation, None)
	pc_tall = PathCondition(None, None, "treating", NodeType.relation, "Faculty")

	pc_cP = PathCondition(pc_c.copy(), None, None, NodeType.plus, None)
	pc_tpc_hr = PathCondition(pc_t.copy(), pc_hr.copy(), None, NodeType.concatenate, None)

	global pc_tpc_hrpc_cP
	pc_tpc_hrpc_cP = PathCondition(pc_tpc_hr.copy(), pc_cP.copy(), None, NodeType.concatenate, None, "Start1", None,
								   "End1", None)

	pc_mpc_tall = PathCondition(pc_m.copy(), pc_tall.copy(), None, NodeType.concatenate, None)
	pc_hrpc_cP = PathCondition(pc_hr.copy(), pc_cP.copy(), None, NodeType.concatenate, None)

	global pc_mpc_tallpc_hrpc_cP
	pc_mpc_tallpc_hrpc_cP = PathCondition(pc_mpc_tall.copy(), pc_hrpc_cP.copy(), None, NodeType.concatenate, None,
										  "Start2", None, "End2", None)

	# This is the Condition of the rule containing 2 path conditions
	# 1: Start1.treating(Faculty);healthRec;contains+.End1
	# 2: Start2.member;treating(Faculty);healthRec;contains+.End2

	global condition
	condition = []
	condition.append(pc_tpc_hrpc_cP)
	condition.append(pc_mpc_tallpc_hrpc_cP)


def setupCondition4():
	# path expresion
	pc_t = PathCondition(None, None, "treating", NodeType.relation, "Faculty,Clinician")
	pc_hr = PathCondition(None, None, "healthRec", NodeType.relation, None)
	pc_c = PathCondition(None, None, "contains", NodeType.relation, None)

	pc_m = PathCondition(None, None, "member", NodeType.relation, None)
	pc_tall = PathCondition(None, None, "treating", NodeType.relation, "Faculty,Item")

	pc_cP = PathCondition(pc_c.copy(), None, None, NodeType.plus, None)
	pc_tpc_hr = PathCondition(pc_t.copy(), pc_hr.copy(), None, NodeType.concatenate, None)

	global pc_tpc_hrpc_cP
	pc_tpc_hrpc_cP = PathCondition(pc_tpc_hr.copy(), pc_cP.copy(), None, NodeType.concatenate, None,  "Clinician",
								   None, "Item", None)

	pc_mpc_tall = PathCondition(pc_m.copy(), pc_tall.copy(), None, NodeType.concatenate, None)
	pc_hrpc_cP = PathCondition(pc_hr.copy(), pc_cP.copy(), None, NodeType.concatenate, None)

	global pc_mpc_tallpc_hrpc_cP
	pc_mpc_tallpc_hrpc_cP = PathCondition(pc_mpc_tall.copy(), pc_hrpc_cP.copy(), None, NodeType.concatenate, None,
										  "Clinician", None, "Item", None)

	# This is the Condition of the rule containing 2 path conditions
	# 1: Clinician . treating(Faculty);healthRec;contains+ . Item
	# 2: Clinician . member;treating(Faculty);healthRec;contains+ . Item

	global condition
	condition = []
	condition.append(pc_tpc_hrpc_cP)
	condition.append(pc_mpc_tallpc_hrpc_cP)


def setupCondition5():
	# path expresion
	pc_t = PathCondition(None, None, "treating", NodeType.relation, "Faculty")
	pc_hr = PathCondition(None, None, "healthRec", NodeType.relation, None)
	pc_c = PathCondition(None, None, "contains", NodeType.relation, None)

	pc_m = PathCondition(None, None, "member", NodeType.relation, None)
	pc_mR = PathCondition(pc_m.copy(), None, None, NodeType.reverse, None)
	pc_tall = PathCondition(None, None, "treating", NodeType.relation, "Faculty")

	pc_cP = PathCondition(pc_c.copy(), None, None, NodeType.plus, None)
	pc_tpc_hr = PathCondition(pc_t.copy(), pc_hr.copy(), None, NodeType.concatenate, None)

	global pc_tpc_hrpc_cP
	pc_tpc_hrpc_cP = PathCondition(pc_tpc_hr.copy(), pc_cP.copy(), None, NodeType.concatenate, None,  "Clinician",
								   None, "Item", None)

	pc_mRpc_tall = PathCondition(pc_mR.copy(), pc_tall.copy(), None, NodeType.concatenate, None)
	pc_hrpc_cP = PathCondition(pc_hr.copy(), pc_cP.copy(), None, NodeType.concatenate, None)

	global pc_mpc_tallpc_hrpc_cP
	pc_mpc_tallpc_hrpc_cP = PathCondition(pc_mRpc_tall.copy(), pc_hrpc_cP.copy(), None, NodeType.concatenate,
										  None, "Clinician", None, "Item", None)

	# This is the Condition of the rule containing 2 path conditions
	# 1: Clinician . treating(Faculty);healthRec;contains+ . Item
	# 2: Clinician . member(reverese);treating(Faculty);healthRec;contains+ . Item

	global condition
	condition = []
	condition.append(pc_tpc_hrpc_cP)
	condition.append(pc_mpc_tallpc_hrpc_cP)


def setupCondition6():
	# path expresion
	pc_t = PathCondition(None, None, "treating", NodeType.relation, "Faculty")
	pc_hr = PathCondition(None, None, "healthRec", NodeType.relation, None)
	pc_c = PathCondition(None, None, "contains", NodeType.relation, "Faculty")

	pc_m = PathCondition(None, None, "member", NodeType.relation, None)
	pc_tall = PathCondition(None, None, "treating", NodeType.relation, "Faculty")

	pc_cP = PathCondition(pc_c.copy(), None, None, NodeType.plus, "Faculty")
	pc_tpc_hr = PathCondition(pc_t.copy(), pc_hr.copy(), None, NodeType.concatenate, None)

	global pc_tpc_hrpc_cP
	pc_tpc_hrpc_cP = PathCondition(pc_tpc_hr.copy(), pc_cP.copy(), None, NodeType.concatenate, None, "Clinician",
								   None, "Item", None)

	pc_mpc_tall = PathCondition(pc_m.copy(), pc_tall.copy(), None, NodeType.concatenate, None)
	pc_hrpc_cP = PathCondition(pc_hr.copy(), pc_cP.copy(), None, NodeType.concatenate, None)

	global pc_mpc_tallpc_hrpc_cP
	pc_mpc_tallpc_hrpc_cP = PathCondition(pc_mpc_tall.copy(), pc_hrpc_cP.copy(), None, NodeType.concatenate, None,
										  "Clinician", None, "Item", None)

	# This is the Condition of the rule containing 2 path conditions
	# 1: Clinician . treating(Faculty);healthRec;contains+ . Item
	# 2: Clinician . member;treating(Faculty);healthRec;contains+ . Item

	global condition
	condition = []
	condition.append(pc_tpc_hrpc_cP)
	condition.append(pc_mpc_tallpc_hrpc_cP)


def setupCondition7():
	# path expresion
	pc_t = PathCondition(None, None, "treating", NodeType.relation, "*")
	pc_hr = PathCondition(None, None, "healthRec", NodeType.relation, None)
	pc_c = PathCondition(None, None, "contains", NodeType.relation, None)

	pc_m = PathCondition(None, None, "member", NodeType.relation, None)
	pc_tall = PathCondition(None, None, "treating", NodeType.relation, "*")

	pc_cP = PathCondition(pc_c.copy(), None, None, NodeType.plus, None)
	pc_tpc_hr = PathCondition(pc_t.copy(), pc_hr.copy(), None, NodeType.concatenate, None)

	global pc_tpc_hrpc_cP
	pc_tpc_hrpc_cP = PathCondition(pc_tpc_hr.copy(), pc_cP.copy(), None, NodeType.concatenate, None, "Clinician",
								   None, "Item", None)

	pc_mpc_tall = PathCondition(pc_m.copy(), pc_tall.copy(), None, NodeType.concatenate, None)
	pc_hrpc_cP = PathCondition(pc_hr.copy(), pc_cP.copy(), None, NodeType.concatenate, None)

	global pc_mpc_tallpc_hrpc_cP
	pc_mpc_tallpc_hrpc_cP = PathCondition(pc_mpc_tall.copy(), pc_hrpc_cP.copy(), None, NodeType.concatenate, None,
										  "Clinician", None, "Item", None)

	# This is the Condition of the rule containing 2 path conditions
	# 1: Clinician . treating(*);healthRec;contains+ . Item
	# 2: Clinician . member;treating(*);healthRec;contains+ . Item

	global condition
	condition = []
	condition.append(pc_tpc_hrpc_cP)
	condition.append(pc_mpc_tallpc_hrpc_cP)


def setupCondition8():
	# path expresion
	pc_t = PathCondition(None, None, "treating", NodeType.relation, "Faculty")
	pc_hr = PathCondition(None, None, "healthRec", NodeType.relation, None)
	pc_c = PathCondition(None, None, "contains", NodeType.relation, None)

	pc_m = PathCondition(None, None, "member", NodeType.relation, None)
	pc_tall = PathCondition(None, None, "treating", NodeType.relation, "Faculty")

	pc_cP = PathCondition(pc_c.copy(), None, None, NodeType.plus, None)
	pc_tpc_hr = PathCondition(pc_t.copy(), pc_hr.copy(), None, NodeType.concatenate, None)

	global pc_tpc_hrpc_cP
	pc_tpc_hrpc_cP = PathCondition(pc_tpc_hr.copy(), pc_cP.copy(), None, NodeType.concatenate, None, "Clinician",
								   None, "It", None)

	pc_mpc_tall = PathCondition(pc_m.copy(), pc_tall.copy(), None, NodeType.concatenate, None)
	pc_hrpc_cP = PathCondition(pc_hr.copy(), pc_cP.copy(), None, NodeType.concatenate, None)

	global pc_mpc_tallpc_hrpc_cP
	pc_mpc_tallpc_hrpc_cP = PathCondition(pc_mpc_tall.copy(), pc_hrpc_cP.copy(), None, NodeType.concatenate, None,
										  "Clinician", None, "Item", None)

	# This is the Condition of the rule containing 2 path conditions
	# 1: Clinician . treating(Faculty);healthRec;contains+ . Item
	# 2: Clinician . member;treating(Faculty);healthRec;contains+ . Item

	global condition
	condition = []
	condition.append(pc_tpc_hrpc_cP)
	condition.append(pc_mpc_tallpc_hrpc_cP)


def setupCondition9():
	# path expresion
	pc_t = PathCondition(None, None, "treating", NodeType.relation, None, "Babulnath")
	pc_hr = PathCondition(None, None, "healthRec", NodeType.relation, None)
	pc_c = PathCondition(None, None, "contains", NodeType.relation, None)

	pc_m = PathCondition(None, None, "member", NodeType.relation, None)
	pc_tall = PathCondition(None, None, "treating", NodeType.relation, "Faculty", "Babulnath")

	pc_cP = PathCondition(pc_c.copy(), None, None, NodeType.plus, None)
	pc_tpc_hr = PathCondition(pc_t.copy(), pc_hr.copy(), None, NodeType.concatenate, None)

	global pc_tpc_hrpc_cP
	pc_tpc_hrpc_cP = PathCondition(pc_tpc_hr.copy(), pc_cP.copy(), None, NodeType.concatenate, None, "Clinician",
								   None, "Item", None)

	pc_mpc_tall = PathCondition(pc_m.copy(), pc_tall.copy(), None, NodeType.concatenate, None)
	pc_hrpc_cP = PathCondition(pc_hr.copy(), pc_cP.copy(), None, NodeType.concatenate, None)

	global pc_mpc_tallpc_hrpc_cP
	pc_mpc_tallpc_hrpc_cP = PathCondition(pc_mpc_tall.copy(), pc_hrpc_cP.copy(), None, NodeType.concatenate, None,
										  "Clinician", None, "Item", None)

	# This is the Condition of the rule containing 2 path conditions
	# 1: Clinician . treating(Faculty);healthRec;contains+ . Item
	# 2: Clinician . member;treating(Faculty);healthRec;contains+ . Item

	global condition
	condition = []
	condition.append(pc_tpc_hrpc_cP)
	condition.append(pc_mpc_tallpc_hrpc_cP)


def setupCondition10():
	# path expresion
	pc_p = PathCondition(None, None, "p", NodeType.relation, "X,A" )
	pc_q = PathCondition(None, None, "q", NodeType.relation, "X,B,C")
	pc_r = PathCondition(None, None, "r", NodeType.relation, "B")

	pc_ppc_q = PathCondition(pc_p.copy(), pc_q.copy(), None, NodeType.concatenate, None)
	pc_rpc_r = PathCondition(pc_r.copy(), pc_r.copy(), None, NodeType.concatenate, None)


	global pc_ppc_qpc_rpc_r
	pc_ppc_qpc_rpc_r = PathCondition(pc_ppc_q.copy(), pc_rpc_r.copy(), None, NodeType.concatenate, None,  "S",
								   None, "T", None)


	global condition
	condition = []
	condition.append(pc_ppc_qpc_rpc_r)

def setupCondition11():
	# path expresion
	pc_p = PathCondition(None, None, "p", NodeType.relation, "X,A" )
	pc_q = PathCondition(None, None, "q", NodeType.relation, "X,B,B")
	pc_r = PathCondition(None, None, "r", NodeType.relation, "B")

	pc_ppc_q = PathCondition(pc_p.copy(), pc_q.copy(), None, NodeType.concatenate, None)
	pc_rpc_r = PathCondition(pc_r.copy(), pc_r.copy(), None, NodeType.concatenate, None)


	global pc_ppc_qpc_rpc_r
	pc_ppc_qpc_rpc_r = PathCondition(pc_ppc_q.copy(), pc_rpc_r.copy(), None, NodeType.concatenate, None,  "S",
								   None, "T", None)


	global condition
	condition = []
	condition.append(pc_ppc_qpc_rpc_r)

def setupCondition12():
	# path expresion
	pc_p = PathCondition(None, None, "p", NodeType.relation, "X,A" )
	pc_q = PathCondition(None, None, "q", NodeType.relation, "X,B,C")
	pc_rb = PathCondition(None, None, "r", NodeType.relation, "B")
	pc_rc = PathCondition(None, None, "r", NodeType.relation, "C")

	pc_ppc_q = PathCondition(pc_p.copy(), pc_q.copy(), None, NodeType.concatenate, None)
	pc_rpc_r = PathCondition(pc_rb.copy(), pc_rc.copy(), None, NodeType.concatenate, None)


	global pc_ppc_qpc_rpc_r
	pc_ppc_qpc_rpc_r = PathCondition(pc_ppc_q.copy(), pc_rpc_r.copy(), None, NodeType.concatenate, None,  "S",
								   None, "T", None)


	global condition
	condition = []
	condition.append(pc_ppc_qpc_rpc_r)

def setupCondition13():
	# path expresion
	# path expresion
	pc_p = PathCondition(None, None, "p", NodeType.relation, "X,A")
	pc_q = PathCondition(None, None, "q", NodeType.relation, "A,B,4")
	pc_r = PathCondition(None, None, "r", NodeType.relation, "B")


	pc_ppc_q = PathCondition(pc_p.copy(), pc_q.copy(), None, NodeType.concatenate, None)
	pc_rpc_r = PathCondition(pc_r.copy(), pc_r.copy(), None, NodeType.concatenate, None)


	global pc_ppc_qpc_rpc_r
	pc_ppc_qpc_rpc_r = PathCondition(pc_ppc_q.copy(), pc_rpc_r.copy(), None, NodeType.concatenate, None,  "S",
								   None, "T", None)


	global condition
	condition = []
	condition.append(pc_ppc_qpc_rpc_r)

def setupCondition14():
	# path expresion
	# path expresion
	pc_p = PathCondition(None, None, "p", NodeType.relation, "X,A")
	pc_q = PathCondition(None, None, "q", NodeType.relation, "X,B,2")
	pc_r = PathCondition(None, None, "r", NodeType.relation, "B")


	pc_ppc_q = PathCondition(pc_p.copy(), pc_q.copy(), None, NodeType.concatenate, None)
	pc_rpc_r = PathCondition(pc_r.copy(), pc_r.copy(), None, NodeType.concatenate, None)


	global pc_ppc_qpc_rpc_r
	pc_ppc_qpc_rpc_r = PathCondition(pc_ppc_q.copy(), pc_rpc_r.copy(), None, NodeType.concatenate, None,  "S",
								   None, "T", None)

	global condition
	condition = []
	condition.append(pc_ppc_qpc_rpc_r)

def setupCondition15():
	# path expresion
	# path expresion
	pc_p = PathCondition(None, None, "p", NodeType.relation, "A,B")
	pc_ra = PathCondition(None, None, "r", NodeType.relation, "A")
	pc_rb = PathCondition(None, None, "r", NodeType.relation, "B")


	pc_ppc_ra = PathCondition(pc_p.copy(), pc_ra.copy(), None, NodeType.concatenate, None)

	global pc_ppc_rapc_rb
	pc_ppc_rapc_rb = PathCondition(pc_ppc_ra.copy(), pc_rb.copy(), None, NodeType.concatenate, None,  "S",
								   None, "T", None)

	global condition
	condition = []
	condition.append(pc_ppc_rapc_rb)

def setupCondition16():
	# path expresion
	# path expresion

	pc_ra = PathCondition(None, None, "r", NodeType.relation, "A")

	pc_rapc_ra = PathCondition(pc_ra.copy(), pc_ra.copy(), None, NodeType.concatenate, None)

	global pc_rapc_rapc_ra
	pc_rapc_rapc_ra = PathCondition(pc_rapc_ra.copy(), pc_ra.copy(), None, NodeType.concatenate, None,  "S2",
								   None, "T", None)

	global condition
	condition = []
	condition.append(pc_rapc_rapc_ra)

def setupCondition17():
	# path expresion
	# path expresion

	pc_rA = PathCondition(None, None, "r", NodeType.relation, "A")
	pc_r2 = PathCondition(None, None, "r", NodeType.relation, "2")
	pc_rAp =  PathCondition(pc_rA.copy(), None, None, NodeType.plus, None)

	pc_rApc_rAp = PathCondition(pc_rA.copy(), pc_rAp.copy(), None, NodeType.concatenate, None)

	pc_p = PathCondition(None, None, "r", NodeType.relation, "B,4")
	pc_rApc_rAppc_p = PathCondition(pc_rApc_rAp.copy(), pc_p.copy(), None, NodeType.concatenate, None,  "S2",
								   None, "T", None)

	global condition
	condition = []
	condition.append(pc_rApc_rAppc_p)

def setupCondition18():
	# path expresion
	# path expresion

	pc_rA = PathCondition(None, None, "r", NodeType.relation, "A")
	pc_r2 = PathCondition(None, None, "r", NodeType.relation, "2")
	pc_rAp =  PathCondition(pc_rA.copy(), None, None, NodeType.plus, None)

	pc_r2pc_rAp = PathCondition(pc_r2.copy(), pc_rAp.copy(), None, NodeType.concatenate, None)

	pc_p = PathCondition(None, None, "p", NodeType.relation,"B,4")
	pc_r2pc_rAppc_p = PathCondition(pc_r2pc_rAp.copy(),pc_p.copy(),None, NodeType.concatenate, None,  "S2",
								   None, "T", None)

	global condition
	condition = []
	condition.append(pc_r2pc_rAppc_p)

def setupCondition19():
	# path expresion
	# path expresion

	pc_rA = PathCondition(None, None, "r", NodeType.relation, "A")
	pc_r2 = PathCondition(None, None, "r", NodeType.relation, "2")
	pc_rAs =  PathCondition(pc_rA.copy(), None, None, NodeType.star, None)

	pc_r2pc_rAs = PathCondition(pc_r2.copy(), pc_rAs.copy(), None, NodeType.concatenate, None)

	pc_p = PathCondition(None, None, "p", NodeType.relation,"B,4")
	pc_r2sc_rAppc_p = PathCondition(pc_r2pc_rAs.copy(),pc_p.copy(),None, NodeType.concatenate, None,  "S2",
								   None, "T", None)

	global condition
	condition = []
	condition.append(pc_r2sc_rAppc_p)

def setupCondition20():
	# path expresion
	# path expresion

	pc_rA = PathCondition(None, None, "r", NodeType.relation, "A")
	pc_r2 = PathCondition(None, None, "r", NodeType.relation, "2")
	pc_rAp =  PathCondition(pc_rA.copy(), None, None, NodeType.plus, None)

	pc_r2pc_rAp = PathCondition(pc_r2.copy(), pc_rAp.copy(), None, NodeType.concatenate, None)

	pc_p = PathCondition(None, None, "p", NodeType.relation, "B,5")
	pc_ps = PathCondition(pc_p.copy(), None, None, NodeType.star, None)
	pc_r2pc_rAppc_pp = PathCondition(pc_r2pc_rAp.copy(),pc_ps.copy(),None, NodeType.concatenate, None,  "S2",
								   None, "T", None)

	global condition
	condition = []
	condition.append(pc_r2pc_rAppc_pp)

def setupCondition21():
	# path expresion
	# path expresion

	pc_q = PathCondition(None, None, "q", NodeType.relation, "A,3,C")
	pc_p = PathCondition(None, None, "p", NodeType.relation, "3,C")
	pc_pr =  PathCondition(pc_p.copy(), None, None, NodeType.reverse, None)

	pc_qpc_pr = PathCondition(pc_q.copy(), pc_pr.copy(), None, NodeType.concatenate, None,  "S2",
								   None, "T", None)


	global condition
	condition = []
	condition.append(pc_qpc_pr)


def setupSystemGraph1():
	"""System graph shown at the end of pesudocode document"""
	# symmetric relationships
	global s
	s = set()

	# relationships
	global r
	r = s.copy()
	r.add("treating")
	r.add("healthRec")
	r.add("contains")
	r.add("member")

	# graph
	global vertices
	vertices = set()
	vertices.add("Samrat")
	vertices.add("Ishan")
	vertices.add("Ishan_HR")
	vertices.add("Scabies_Item1")
	vertices.add("SamratGroup")

	global edges
	edges = set()
	edges.add(Edge("Samrat", "Ishan", "treating", "Babulnath"))
	edges.add(Edge("Ishan", "Ishan_HR", "healthRec", None))
	edges.add(Edge("Ishan_HR", "Scabies_Item1", "contains", None))

	edges.add(Edge("Samrat", "SamratGroup", "member", None))
	edges.add(Edge("SamratGroup", "Ishan", "treating", "Babulnath"))


def setupSystemGraph2():
	# symmetric relationships
	global s
	s = set()

	# relationships
	global r
	r = s.copy()
	r.add("treating")
	r.add("healthRec")
	r.add("contains")
	r.add("member")

	# graph
	global vertices
	vertices = set()
	vertices.add("Samrat")
	vertices.add("Ishan")
	vertices.add("Ishan_HR")
	vertices.add("Scabies_Item1")
	vertices.add("SamratGroup")

	global edges
	edges = set()
	edges.add(Edge("Samrat", "Ishan", "treating", "Babulnath"))
	edges.add(Edge("Ishan", "Ishan_HR", "healthRec", None))
	edges.add(Edge("Ishan_HR", "Scabies_Item1", "contains", None))

	edges.add(Edge("Samrat", "SamratGroup", "member", None))
	edges.add(Edge("SamratGroup", "Ishan", "treating", "Hickville"))


def setupSystemGraph3():
	# symmetric relationships
	global s
	s = set()

	# relationships
	global r
	r = s.copy()
	r.add("treating")
	r.add("healthRec")
	r.add("contains")
	r.add("member")

	# graph
	global vertices
	vertices = set()
	vertices.add("Samrat")
	vertices.add("Ishan")
	vertices.add("Ishan_HR")
	vertices.add("Scabies_Item1")
	vertices.add("SamratGroup")

	global edges
	edges = set()
	edges.add(Edge("Samrat", "Ishan", "treating", "Babulnath"))
	edges.add(Edge("Ishan", "Ishan_HR", "healthRec", None))
	edges.add(Edge("Ishan_HR", "Scabies_Item1", "contains", None))

	edges.add(Edge("Samrat", "SamratGroup", "member", None))
	edges.add(Edge("SamratGroup", "Ishan", "treating", "Hickville"))


def setupSystemGraph4():
	# symmetric relationships
	global s
	s = set()

	# relationships
	global r
	r = s.copy()
	r.add("treating")
	r.add("healthRec")
	r.add("contains")
	r.add("member")

	# graph
	global vertices
	vertices = set()
	vertices.add("Samrat")
	vertices.add("Ishan")
	vertices.add("Ishan_HR")
	vertices.add("Scabies_Item1")
	vertices.add("SamratGroup")

	global edges
	edges = set()
	edges.add(Edge("Samrat", "Ishan", "treating", "Babulnath,Samrat"))
	edges.add(Edge("Ishan", "Ishan_HR", "healthRec", None))
	edges.add(Edge("Ishan_HR", "Scabies_Item1", "contains", None))

	edges.add(Edge("Samrat", "SamratGroup", "member", None))
	edges.add(Edge("SamratGroup", "Ishan", "treating", "Babulnath,Scabies_Item1"))


def setupSystemGraph5():
	# symmetric relationships
	global s
	s = set()

	# relationships
	global r
	r = s.copy()
	r.add("treating")
	r.add("healthRec")
	r.add("contains")
	r.add("member")

	# graph
	global vertices
	vertices = set()
	vertices.add("Samrat")
	vertices.add("Ishan")
	vertices.add("Ishan_HR")
	vertices.add("Scabies_Item1")
	vertices.add("SamratGroup")

	global edges
	edges = set()
	edges.add(Edge("Samrat", "Ishan", "treating", "Babulnath"))
	edges.add(Edge("Ishan", "Ishan_HR", "healthRec", None))
	edges.add(Edge("Ishan_HR", "Scabies_Item1", "contains", None))

	edges.add(Edge("SamratGroup", "Samrat", "member", None))
	edges.add(Edge("SamratGroup", "Ishan", "treating", "Babulnath"))


def setupSystemGraph6():
	# symmetric relationships
	global s
	s = set()

	# relationships
	global r
	r = s.copy()
	r.add("treating")
	r.add("healthRec")
	r.add("contains")
	r.add("member")

	# graph
	global vertices
	vertices = set()
	vertices.add("Samrat")
	vertices.add("Ishan")
	vertices.add("Ishan_HR")
	vertices.add("Scabies_Item1")
	vertices.add("SamratGroup")

	global edges
	edges = set()
	edges.add(Edge("Samrat", "Ishan", "treating", "Babulnath"))
	edges.add(Edge("Ishan", "Ishan_HR", "healthRec", None))
	edges.add(Edge("Ishan_HR", "Scabies_Item1", "contains", "Babulnath"))
	edges.add(Edge("Scabies_Item1", "Scabies_Med_Rep", "contains", "Babulnath"))

	edges.add(Edge("Samrat", "SamratGroup", "member", None))
	edges.add(Edge("SamratGroup", "Ishan", "treating", "Babulnath"))


def setupSystemGraph7():
	# symmetric relationships
	global s
	s = set()

	# relationships
	global r
	r = s.copy()
	r.add("treating")
	r.add("healthRec")
	r.add("contains")
	r.add("member")

	# graph
	global vertices
	vertices = set()
	vertices.add("Samrat")
	vertices.add("Ishan")
	vertices.add("Ishan_HR")
	vertices.add("Scabies_Item1")
	vertices.add("SamratGroup")

	global edges
	edges = set()
	edges.add(Edge("Samrat", "Ishan", "treating", "Babulnath"))
	edges.add(Edge("Ishan", "Ishan_HR", "healthRec", None))
	edges.add(Edge("Ishan_HR", "Scabies_Item1", "contains", None))

	edges.add(Edge("Samrat", "SamratGroup", "member", None))
	edges.add(Edge("SamratGroup", "Ishan", "treating", None))


def setupSystemGraph8():
	# symmetric relationships
	global s
	s = set()

	# relationships
	global r
	r = s.copy()
	r.add("treating")
	r.add("healthRec")
	r.add("contains")
	r.add("member")

	# graph
	global vertices
	vertices = set()
	vertices.add("Samrat")
	vertices.add("Ishan")
	vertices.add("Ishan_HR")
	vertices.add("Scabies_Item1")
	vertices.add("SamratGroup")

	global edges
	edges = set()
	edges.add(Edge("Samrat", "Ishan", "treating", "Babulnath"))
	edges.add(Edge("Ishan", "Ishan_HR", "healthRec", None))
	edges.add(Edge("Ishan_HR", "Scabies_Item1", "contains", None))

	edges.add(Edge("Samrat", "SamratGroup", "member", None))
	edges.add(Edge("SamratGroup", "Ishan", "treating", "Babulnath"))
	edges.add(Edge("SamratGroup", "Ishan", "treating", "Babulnath"))


def setupSystemGraph9():
	"""System graph shown at the end of pesudocode document"""
	# symmetric relationships
	global s
	s = set()

	# relationships
	global r
	r = s.copy()
	r.add("treating")
	r.add("healthRec")
	r.add("contains")
	r.add("member")

	# graph
	global vertices
	vertices = set()
	vertices.add("Samrat")
	vertices.add("Ishan")
	vertices.add("Ishan_HR")
	vertices.add("Scabies_Item1")
	vertices.add("SamratGroup")

	global edges
	edges = set()
	edges.add(Edge("Samrat", "Ishan", "treating", "Babulnath"))
	edges.add(Edge("Ishan", "Ishan_HR", "healthRec", None))
	edges.add(Edge("Ishan_HR", "Scabies_Item1", "contains", None))

	edges.add(Edge("Samrat", "SamratGroup", "member", None))
	edges.add(Edge("SamratGroup", "Ishan", "treating", "Babulnath"))

def setupSystemGraph10():
	"""System graph shown at the end of pesudocode document"""
	# symmetric relationships
	global s
	s = set()

	# relationships
	global r
	r = s.copy()
	r.add("p")
	r.add("q")
	r.add("r")

	# graph
	global vertices
	vertices = set()
	vertices.add("a")
	vertices.add("b1")
	vertices.add("b2")
	vertices.add("c1")
	vertices.add("c2")

	global edges
	edges = set()
	edges.add(Edge("a", "b1", "p", "1,2"))
	edges.add(Edge("b1", "c1", "q", "1,2,3"))
	edges.add(Edge("b1", "b2", "q", "2,3,4"))
	edges.add(Edge("b2", "c2", "r", "3"))
	edges.add(Edge("a", "b2", "p", "3,4"))
	edges.add(Edge("c2", "a", "r", "3"))
	edges.add(Edge("c2", "c1", "r", "4"))
	edges.add(Edge("c1", "c2", "r", "3"))
	edges.add(Edge("c1", "c1", "r", "2"))

def evaluateRule(r, req):
	m = getInitialMapping(r, req)
	#parent is a map of parent pointers.  for a tuple t added to SEEN during the BFS for path condition i, parent((i,t)) is the tuple from which t was first reached during that BFS.
	parent = {}
	satisfyingTuple = []
	return (evaluatePathCondition(0, m, parent, satisfyingTuple))

def getInitialMapping(r, req):
	mapping = {}
	mapping.update({r.subject: req.user})
	rule_action_resource = r.action.resource.split(",")
	req_resource = req.resource.split(",")
	index = 0 
	for res in rule_action_resource:
		mapping.update({res: req_resource[index]})
		index = index +1
	print (str(mapping))
	return mapping

# ================================================

def getStartNodes(index, m):
	if condition[index].start_constant is not None:
		return [condition[index].start_constant]
	elif condition[index].start_variable is not None:
		for parameter in m.keys():
			if (parameter == condition[index].start_variable):
				return [m[parameter]]

	return vertices


def getDestinationNodes(index, m):
	if condition[index].dest_constant is not None:
		return [condition[index].dest_constant]
	elif condition[index].dest_variable is not None:
		for parameter in m.keys():
			if (parameter == condition[index].dest_variable):
				return [m[parameter]]

	return vertices


def evaluatePathCondition(index, m, parent, satisfyingTuple):
	# Get the start node from the path condition and the mapping. If a mapping does not exist a list of all the nodes will be sent
	startNodes = getStartNodes(index, m)

	# Get the destination node from the path condition and the mapping. If a mapping does not exist a list of all the nodes will be sent
	destNodes = getDestinationNodes(index, m)

	for src in startNodes:
		if condition[index].start_variable is not None:
			m.update({condition[index].start_variable:src})
		global pi
		pi = condition[index]
		if (evaluatePathConditionGivenIndex(index, m, src, destNodes, parent, satisfyingTuple) == True):
			return True
	return False


def evaluateParameters(m, para, p, new_map):
	index = 0
	parameterMatchFailed = False
	for par in para:
		if ((par[0].isupper())):
			for k, v in m.items():
				par_exist = False
				if (k == par):
					par_exist = True
					if (v != p[index]):
						parameterMatchFailed = True
						break
			if par_exist == False:
				new_map.update({par: p[index]})
		else:
			if (par != p[index]):
				parameterMatchFailed = True
		if parameterMatchFailed == True:
			break
		index = index + 1
	return parameterMatchFailed

def getPath(i,node, parent):
	path = [node]
	while parent.get((i,node)) is not None:
		node = parent.get((i,node))
		path.append(node)

	path.reverse()
	nodePath = []
	for i in range(len(path)):
		nodePath.append(path[i].node)

	return nodePath

def evaluatePathConditionGivenIndex(index, m, src, destNodes, parent, satisfyingTuple):
	"""MatchPrincipal Algorithm"""
	print("Evaluation Algorithm")
	print("Please note that for ease of reading the symbol ¬ is used to indicate \overline.")

	print("\n\tSubject Node: " + src)
	print("\tDestination Node: " + str(destNodes))
	print("\tPath Condition: " + pi.to_string() + "\n")

	# Initialize empty queue Q
	Q = queue.Queue()
	# Initialize empty set of visited nodes SEEN

	SEEN = OrderedSet()
	# add (u, π) to Q
	# Each tuple of the Q contains a node to be visited, the path condition it needs to be evaluated against and the mapping of parameters seen till now.
	Q.put(VisitedNode(src, pi, m))
	# SEEN = (u, π) ∪ SEEN
	# The SEEN list contains the tuples that are visited. Each tuple contains a node which was visited, the path condition that node was evaluated against, and the mapping of the parameters.
	SEEN.add(VisitedNode(src, pi, m))

	# while Q is not empty do
	while not Q.empty():
		print("\tQ has {0:d} item(s).".format(Q.qsize()))
		# dequeue next entry (h, φ) from Q
		hphi = Q.get()
		print("\n\tVisiting: " + hphi.node + ", with path: " + hphi.pathCondition.to_string())
		# Initialize empty list of node, suffix tuples Θ
		Theta = []

		# ====== debug start ======
		for seenNode in SEEN:
			debugprint(2, "- Seen Already: " + seenNode.node + " with path: " + seenNode.pathCondition.to_string())
		# ====== debug end ======
		# for each edge (h, w, r, p) ∈ E do
		for edgeUT in edges:
			if edgeUT.srcNode != hphi.node:
				continue
			# ====== debug start ======
			debugprint(0, "")
			debugprint(2, "Edge: " + edgeUT.srcNode + " --> " + edgeUT.dstNode)
			# ====== debug end ======
			# Θ = Θ join DetermineSuffixes(φ, r)
			head = hphi.pathCondition.head()
			if (head[0].parameter_variable is not None):
				print("\n\tParameter Variable:" + str(head[0].parameter_variable))

			# If the relation does not contain any parameter
			if (head[0].parameter_variable is None or head[0].parameter_variable == "*"):
				new_map = copy.deepcopy(hphi.parameterMap)
				Theta.extend(hphi.pathCondition.copy().determine_suffix(
					PathCondition(None, None, edgeUT.rel, NodeType.relation, edgeUT.par), edgeUT.srcNode,
					edgeUT.dstNode, new_map))
			# If the realtion contains a paramter to be evaluated
			else:
				parameter_list = head[0].parameter_variable.split(",")
				parameter_value_list = []
				if (edgeUT.par is not None):
					parameter_value_list = edgeUT.par.split(",")
				if len(parameter_value_list) == len(parameter_list):
					new_map = copy.deepcopy(hphi.parameterMap)
					parameterMatchFailed = evaluateParameters(hphi.parameterMap, parameter_list, parameter_value_list,
															 new_map)
					if (parameterMatchFailed == False):
						Theta.extend(hphi.pathCondition.copy().determine_suffix(
							PathCondition(None, None, edgeUT.rel, NodeType.relation, edgeUT.par), edgeUT.srcNode,
							edgeUT.dstNode, new_map))


			# if (r ∈ S and (w, h, r, p) !∈ E) then
			if edgeUT.rel in s and Edge(edgeUT.dstNode, edgeUT.srcNode, edgeUT.rel) not in edges:
				# Θ = Θ join DetermineSuffixes(φ, \overline{r})
				# If the relation does not contain any parameter
				if (head[0].parameter_variable is None or head[0].parameter_variable == "*"):
					new_map = copy.deepcopy(hphi.parameterMap)
					Theta.extend(hphi.pathCondition.copy().determine_suffix(
						PathCondition(PathCondition(None, None, edgeUT.rel, NodeType.relation), None, None,
									  NodeType.reverse), edgeUT.srcNode, edgeUT.dstNode, new_map))
				# If the relation contains a parameter to be evaluated
				else:
					parameter_list = head[0].parameter_variable.split(",")
					parameter_value_list = []
					if (edgeUT.par is not None):
						parameter_value_list = edgeUT.par.split(",")
					if len(parameter_value_list) == len(parameter_list):
						new_map = copy.deepcopy(hphi.parameterMap)
						parameterMatchFailed = evaluateParameters(hphi.parameterMap, parameter_list,
																 parameter_value_list, new_map)
						if (parameterMatchFailed == False):
							Theta.extend(hphi.pathCondition.copy().determine_suffix(
								PathCondition(PathCondition(None, None, edgeUT.rel, NodeType.relation), None, None,
											  NodeType.reverse, edgeUT.par), edgeUT.srcNode, edgeUT.dstNode, new_map))

					# end if
		# end for

		# for each edge (w, h, r, p) ∈ E do
		for edgeUT in edges:
			if edgeUT.dstNode != hphi.node:
				continue
			# ====== debug start ======
			debugprint(0, "")
			debugprint(2, "Edge: " + edgeUT.dstNode + " <-- " + edgeUT.srcNode)
			# ====== debug end ======
			# Θ = Θ join DetermineSuffixes(φ, \overline{r})
			head = hphi.pathCondition.head()
			if (head[0].parameter_variable is not None):
				print("\n\tParameter Variable:" + str(head[0].parameter_variable))

			# If the relation does not contain any parameter
			if head[0].parameter_variable is None or head[0].parameter_variable == "*":
				new_map = copy.deepcopy(hphi.parameterMap)
				Theta.extend(hphi.pathCondition.copy().determine_suffix(
					PathCondition(PathCondition(None, None, edgeUT.rel, NodeType.relation, edgeUT.par), None, None,
								  NodeType.reverse, None), edgeUT.dstNode, edgeUT.srcNode, new_map))
			# If the relation contains a paramter to be evaluated
			else:
				parameter_list = head[0].parameter_variable.split(",")
				parameter_value_list = []
				if (edgeUT.par is not None):
					parameter_value_list = edgeUT.par.split(",")
				if len(parameter_value_list) == len(parameter_list):
					new_map = copy.deepcopy(hphi.parameterMap)
					parameterMatchFailed = evaluateParameters(hphi.parameterMap, parameter_list, parameter_value_list,
															 new_map)
					if (parameterMatchFailed == False):
						Theta.extend(hphi.pathCondition.copy().determine_suffix(
							PathCondition(PathCondition(None, None, edgeUT.rel, NodeType.relation, edgeUT.par), None,
										  None,
										  NodeType.reverse, edgeUT.par), edgeUT.dstNode, edgeUT.srcNode, new_map))

			# if (r ∈ S and (h, w, r, p) !∈ E) then
			if edgeUT.rel in s and Edge(edgeUT.dstNode, edgeUT.srcNode, edgeUT.rel) not in edges:
				# Θ = Θ join DetermineSuffixes(φ, r)
				if (head[0].parameter_variable is None or head[0].parameter_variable == "*"):
					new_map = copy.deepcopy(hphi.parameterMap)
					Theta.extend(hphi.pathCondition.copy().determine_suffix(
						PathCondition(None, None, edgeUT.rel, NodeType.relation), edgeUT.dstNode, edgeUT.srcNode),
						new_map)
				else:
					parameter_list = head[0].parameter_variable.split(",")
					parameter_value_list = []
					if (edgeUT.par is not None):
						parameter_value_list = edgeUT.par.split(",")
					if len(parameter_value_list) == len(parameter_list):
						new_map = copy.deepcopy(hphi.parameterMap)
						parameterMatchFailed = evaluateParameters(hphi.parameterMap, parameter_list,
																 parameter_value_list, new_map)
						if (parameterMatchFailed == False):
							Theta.extend(hphi.pathCondition.copy().determine_suffix(
								PathCondition(None, None, edgeUT.rel, NodeType.relation, edgeUT.par), edgeUT.dstNode,
								edgeUT.srcNode),
								new_map)
					# end if
		# end for

		# for each (n, φ s ) ∈ Θ do
		for theta in Theta:
			# if (n, φ s ) !∈ SEEN then
			nextNode = VisitedNode(theta.node, theta.pathCondition, theta.parameterMap)

			seenNodeBefore = False
			for seenNode in SEEN:
				if seenNode.node == nextNode.node and seenNode.pathCondition.to_string() == nextNode.pathCondition.to_string():
					seenNodeBefore = True
					# ====== debug start ======
					debugprint(3,
							   "- Seen Before: " + seenNode.node + " with path: " + seenNode.pathCondition.to_string())
				# ====== debug end ======
			if not seenNodeBefore:
				# if φ s = \diamond then
				if nextNode.pathCondition.relation_label == "\diamond" and nextNode.pathCondition.node_type == NodeType.relation:
					# if n = v then
					if nextNode.node in destNodes and index + 1 == len(condition):
						parent.update({(index,nextNode):hphi})
						satisfyingTuple.insert(index,nextNode)
						if (pi.dest_variable is not None and pi.dest_variable not in nextNode.parameterMap):
							nextNode.parameterMap.update({pi.dest_variable: nextNode.node})
						# return true // match found
						# ====== debug start ======
						debugprint(3,
								   "+ Just Seen: " + nextNode.node + " with path: " + nextNode.pathCondition.to_string())
						debugprint(3,
								   "= Found destination: " + nextNode.node + " with path: " + nextNode.pathCondition.to_string())
						# ====== debug end ======
						print("------------------------------")
						print("Substitution=")
						print("------------------------------")
						for k, v in nextNode.parameterMap.items():
							print(k + ":" + v)
						print("------------------------------")
						for i in range(len(condition)):
							print("Path satisfying conjunct " + str(i) + " = " + str(getPath(i,satisfyingTuple[i], parent)))
						print("------------------------------")
						return True

					# end if
					elif nextNode.node in destNodes:
						# if ()
						if (pi.dest_variable is not None and pi.dest_variable not in nextNode.parameterMap):
							nextNode.parameterMap.update({pi.dest_variable: nextNode.node})

						parent.update({(index,nextNode):hphi})
						satisfyingTuple.insert(index,nextNode)
						if (evaluatePathCondition(index + 1, nextNode.parameterMap, parent, satisfyingTuple) == True):
							return True
				# else
				else:
					# ====== debug start ======
					debugprint(3, "+ Just Seen: " + nextNode.node + " with path: " + nextNode.pathCondition.to_string())
					# ====== debug end ======
					parent.update({(index,nextNode):hphi})
					# add (n, φ s ) to Q
					Q.put(nextNode)
					# SEEN = (n, φ s ) ∪ SEEN
					SEEN.add(nextNode)
				# end if
				# end if
				# end for

	# end while
	# return false // no match
	return False


if __name__ == '__main__':
	main()
	
